%%=========================================================================
%
% GD algorithm for the synchronization of coupled oscillators
%
% Author: Umberto Biccari
% Date: 03/06/2020
%
%%=========================================================================

clear all
clc

%% Input data

N = 10;                           % Total number of oscillators

T0 = 0; T = 3;                    % Time interval
Nt = 100;
tspan = linspace(T0,T,Nt);
dt = tspan(2)-tspan(1);

theta0vec = linspace(0,pi,N);     % Initial datum
theta0 = normpdf(theta0vec);

maxiter = 6000;                   % Maximum number of iterations allowed
beta = 1e-7;                      % Penalization parameter
tol = 1e-4;                       % Tolerance
eta = 10;                         % Lengthstep for GD
u = ones(Nt,1);                   % Initial control

%% Construction of the symbolic model

syms t;                           % Time variable
symTh = sym('theta',[N,1]);       % Direct variable
symP = sym('p',[N,1]);            % Adjoint variable
symOm = sym('om', [N,1]);         % Natural frequency
symU  = sym('u',[1,1]);           % Control
symK = sym('K',[N,N]);            % Coupling

syms Vsys;    
symThth = repmat(symTh,[1,N]);
Vsys = symOm + (symU./N)*sum(symK.*sin(symThth.' - symThth),2);   

jacTh = jacobian(Vsys,symTh);
jacU = jacobian(Vsys,symU);

sigmaOm = 0.1;
OmNum = 0.2*normrnd(0,sigmaOm,N,1);
OmNum = OmNum - mean(OmNum);

KL = abs(max(OmNum)-min(OmNum)) + 1;
KNum = KL*ones(N,N);  
jacU = subs(jacU,symK,KNum);
jacTh = subs(jacTh,symK,KNum);

VsysP = -jacTh.'*symP;
gradJ = jacU.'*symP;

symPsi = norm(sin(symThth.' - symThth),'fro');   
symL = 0.5*beta*(symU.'*symU)';   

symJ = symPsi + symL;
symJFcn = matlabFunction(symJ,'Vars',{symTh,symU});

gradJFcn = matlabFunction(gradJ,'Vars',{symTh,symP});

symF = subs(Vsys,[symOm,symK],[OmNum,KNum]);
symF = vpa(symF,2);
symFFcn = matlabFunction(symF,'Vars',{symTh,symU});
symFadj = subs(VsysP,symK,KNum);
symFadj = vpa(symFadj,2);
symFadjFcn = matlabFunction(symFadj,'Vars',{symTh,symP,symU});

%% Gradient Descent

error = 10;
iter = 0;
p0vec = zeros(N,1);
p0 = 0;
tcost = zeros(1,maxiter);

while (error > tol && iter < maxiter)
    tic
    iter = iter + 1;
    
    % Resolution of the direct dynamics
    thetaOut = zeros(Nt,N);
    thetaOut(1,:) = theta0;
    for m = 1:Nt-1
        thetaFun = @(theta,u) symFFcn(theta,u);  
        thetaOut(m+1,:) = thetaOut(m,:) + dt*thetaFun(thetaOut(m,:)',u(m))';
    end
    
    % Initial datum for the adjoint system
    thetaT = thetaOut(end,:);
    for i = 1:N
        p0vec(i) = p0;
        for j = 1:N
            if j~=i
                p0vec(i) = p0vec(i) + 0.5*sin(2*thetaT(i)-2*thetaT(j));
            end
        end
    end
    
    % Resolution of the adjoint dynamics
    pOut = zeros(Nt,N);
    pOut(end,:) = p0vec;
    
    for m = Nt-1:-1:1
        pFun = @(theta,p,u) symFadjFcn(theta,p,u);
        pOut(m,:) = pOut(m+1,:) - dt*pFun(thetaOut(m+1,:)',pOut(m+1,:)',u(m+1))';
    end
       
    Du = gradJFcn(thetaOut',pOut')';
    
    Du = beta*u + Du;
    uOld = u;
    
    u = u - eta*Du;
    
    DU2 = norm(Du);
    U2 = norm(u);
    Ua2 = norm(uOld-u);
    
    error = DU2/U2;
    
    tcost(iter) = toc;
   
    fprintf("Iteration %i - Error %g\n", iter,error);
end

Tcost = sum(tcost);
%% Resolution of the controlled dynamics

thetaOutOpt = zeros(Nt,N);
thetaOutOpt(1,:) = theta0;
for m = 1:Nt-1
    thetaFunOpt = @(theta,u) symFFcn(theta,u);  
    thetaOutOpt(m+1,:) = thetaOutOpt(m,:) + dt*thetaFunOpt(thetaOut(m,:)',u(m))';
end

%% Resolution of the free dynamics
thetaOutFree = zeros(Nt,N);
thetaOutFree(1,:) = theta0;
for m = 1:Nt-1
    thetaFunFree = @(theta,u) symFFcn(theta,u);  
    thetaOutFree(m+1,:) = thetaOutFree(m,:) + dt*thetaFunOpt(thetaOut(m,:)',1)';
end

%% Plots

close all

figure(1)
for j = 1:N
    plot(thetaOutOpt(:,j),'LineWidth',2)
    hold on
    xlim([1 100])
    xticks([1 34 67 100])
    xticklabels({'0','1','2','3'})
    yticks([])
    a = get(gca,'XTickLabel');
    set(gca,'XTickLabel',a,'fontsize',16)
    xlabel('Time (s)')
    ylabel('Oscillators')    
end

set(gcf, 'Units', 'Normalized', 'OuterPosition', [0, 0.04, 0.4, 0.4])

%%
figure(2)
for j = 1:N
    plot(thetaOutFree(:,j),'LineWidth',2)
    hold on
    xlim([1 100])
    xticks([1 34 67 100])
    xticklabels({'0','1','2','3'})
    yticks([])
    a = get(gca,'XTickLabel');
    set(gca,'XTickLabel',a,'fontsize',16)
    xlabel('Time (s)')
    ylabel('Oscillators')    
end

set(gcf, 'Units', 'Normalized', 'OuterPosition', [0, 0.04, 0.4, 0.4])

%%
figure(3)
plot(u,'LineWidth',2)
xlim([1 100])
xticks([1 34 67 100])
xticklabels({'0','1','2','3'})
a = get(gca,'XTickLabel');
set(gca,'XTickLabel',a,'fontsize',16)
xlabel('Time (s)')
Leg = legend('Control');
Leg.Location = 'northwest';

set(gcf, 'Units', 'Normalized', 'OuterPosition', [0, 0.04, 0.4, 0.4])